// $Id: file.c,v 1.11 2005/06/17 04:13:17 lundman Exp $
// File IO functions for the lion library
// Jorgen Lundman January 8th, 2003.


// Universal system headers

#ifndef WIN32

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <signal.h>

#include <unistd.h>

#include <fcntl.h>   // for locking



#ifdef __sun__
extern const char *const sys_errlist[];
#endif



// Implementation specific headers

#include "connections.h"
#include "io.h"
#include "sockets.h"

#include "lion.h"




__RCSID("$LiON: lundman/lion/file.c,v 1.11 2005/06/17 04:13:17 lundman Exp $");



connection_t *lion_open(char *file, int flags, mode_t modes,
						int lion_flags, void *user_data)
{
	connection_t *newd;

	// This breaks FULLFILL, but then, passing NULL breaks lion API.
	if (!file)
		return NULL;
	
	newd = connections_new();
	
	newd->type = LION_TYPE_FILE;
	newd->user_data = user_data;

	if (lion_flags & LION_FLAG_TRACE)
		lion_enable_trace(newd);

	if (newd->trace)
		fprintf(trace_file, "%p: lion_open (%s)\n",newd, file);

	// Open the file, if issues, send event.

	if (lion_flags & LION_FLAG_EXCLUSIVE) {



		// They want exclusive, so lets step through the various ways this
		// can be done, in descending order of preference.
		// FYI, NetBSD has all of these, which made it easier to test.
		
#ifdef O_EXLOCK

		newd->socket = open( file, flags | O_EXLOCK | O_NONBLOCK, modes );

#elif defined LOCK_EX
		
		newd->socket = open( file, flags, modes );

		if (newd->socket >= 0) {
			if (flock(newd->socket, LOCK_EX|LOCK_NB)) {
				close(newd->socket);
				newd->socket = -1;
			}
		}

#elif defined F_SETLK

		{
			struct flock locker;

			locker.l_start = 0;         // offset from whence
			locker.l_len   = 0;         // whole file
			locker.l_pid   = getpid();  // don't actually care
			locker.l_type  = F_WRLCK;
			locker.l_whence= 0;         // start of file.

			newd->socket = open( file, flags, modes );
			
			if (newd->socket >= 0) {

				if (fcntl( newd->socket, F_SETLK, &locker) == -1) {

					close(newd->socket);
					newd->socket = -1;

				}
			}
		}

#elif defined F_TLOCK  // advisory only, but atleast we can stop within lion.

		newd->socket = open( file, flags, modes );
		
		if (newd->socket >= 0) {
			
			if ( lockf( newd->socket, F_TLOCK, 0) == -1) {
				
				close(newd->socket);
				newd->socket = -1;

			}
		}


#endif  // which locking type

		
		// Phew no exclusive, just open.

	} else {
		
		newd->socket = open( file, flags, modes );

	}

#ifdef DEBUG
	if (newd->socket < 0)
		perror("open");
	printf("net_open(%s): %d\n", file, newd->socket);
#endif

	if (newd->socket < 0) {

		if (lion_flags & LION_FLAG_FULFILL) {

			// Signal above layer of failure
			newd->status = ST_PENDING;

			io_force_loop = 1;
			return newd;
		}
	
		newd->status = ST_DISCONNECT;

		return NULL;
	}


	// File was opened ok, set nonblocking.
	// Technically not needed if the first locking method exist.
	
	// In Unix, we can just use the same method as sockets
	//fcntl(newd->socket,F_SETFL,(fcntl(newd->socket,F_GETFL)|O_NONBLOCK));
	newd->socket = sockets_setnonblocking(newd->socket);


	//	printf("[file] file %d is nonblocking %d\n", newd->socket,
	//	   fcntl(newd->socket, F_GETFL) & O_NONBLOCK);

	
	// We're done, tell application its ready.
	// Race condition here, we can signal app that it is CONNECTED but
	// since we've not returned the handle, the client will get rather
	// confused.

	newd->status = ST_PENDING;
	
	return newd;

}




#endif
