// $Id: file_win32.c,v 1.13 2005/05/25 12:33:38 lundman Exp $
// File IO functions for the lion library
// Jorgen Lundman January 8th, 2003.


// Universal system headers
#ifdef WIN32


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <signal.h>

#include <io.h>
#include <direct.h>
#include <share.h>



// Implementation specific headers

#include "connections.h"
#include "io.h"
#include "lion.h"
#include "sockets.h"

#include "lion_rcsid.h"


__RCSID("$LiON: lundman/lion/file_win32.c,v 1.13 2005/05/25 12:33:38 lundman Exp $");



//
// This is a thread handler to read file IO.
//
// Once started it will wait for a mutex, which is the parent telling
// it when it is allowed to start.
// Then it will enter a loop reading from a file, writing to the pipe.
//
int lion_filereader(void *arg)
{
	connection_t *node;
	int temp;
	int red = 0, wrote = 0, ret = 0;

#define THREAD_IO_SIZE node->inbuffer_size

	node = (connection_t *) arg;

	if (!node) 
			goto all_done;


#ifdef DEBUG
	printf("[lion_filereader] waiting permission to start!\n");
#endif

	WaitForSingleObject( node->mutex, INFINITE );
	//CloseHandle( node->mutex );
	node->mutex = NULL;

#ifdef DEBUG
	printf("[lion_filereader] Starting..\n");
#endif

	// Kick pipe into blocking IO again..
	temp = 0;
	ioctlsocket(node->socket, FIONBIO, &temp);



	while(  ( red = read(node->file_socket, node->buffer, THREAD_IO_SIZE) ) > 0) {

		wrote = sockets_write( node->socket, node->buffer, red );

		lion_poll(0, 0);  // the read is blocking, so just go as fast as possible!
 		//printf("   io %d %d -> %d\n", red, wrote, node->socket);

		if (wrote != red) {
			break;
		}
	}



all_done:

	if (red == -1) // Error, signal failure.
		ret = errno;
	if (wrote == -1)
		ret = WSAGetLastError();

	// Ok, we need to cleanup. Nobody closes the file
	// so we have to here.
	close(node->file_socket);
	node->file_socket = -1;
#ifdef DEBUG
	printf("[file_win32] closed the file: status %d\n", ret);
#endif

	// The pipe will be closed, by master.
	node->status = ST_DISCONNECT;
	node->type = LION_TYPE_PIPE_FILE; // Is there a danger master see's this, before we get to exit?

	//printf("[lion_filereader] signalling %p to be released\n", node);

	//io_force_loop = 1;
	// Alas, we can not to force loop here, since that is OUR force loop, not parents.
	// So, we close the connection instead, this should wake the parent up.
	temp = node->socket;
	node->socket = -1; // Avoid the race condition.
	sockets_close(temp);

	return ret;
}



int lion_filewriter(void *arg)
{
	connection_t *node;
	int temp;
	int red = 0, wrote = 0, ret = 0;


	node = (connection_t *) arg;


	if (!node) 
			goto all_done;

	WaitForSingleObject( node->mutex, INFINITE );
	CloseHandle( node->mutex );
	node->mutex = NULL;


	//printf("[lion_filewriter] Starting..\n");

	// Kick pipe into blocking IO again..
	temp = 0;
	ioctlsocket(node->socket, FIONBIO, &temp);



	while(  ( red = sockets_read(node->socket, node->buffer, THREAD_IO_SIZE) ) > 0) {

		wrote = write( node->file_socket, node->buffer, red );

		//printf("   io %d %d -> %d\n", red, wrote, node->socket);

		if (wrote != red) {
			break;
		}
	}



all_done:

	if (red == -1) // Error, signal failure.
		ret = WSAGetLastError();
	if (wrote == -1)
		ret = errno;

	// Ok, we need to cleanup. Nobody closes the file
	// so we have to here.
	close(node->file_socket);
	node->file_socket = -1;
#ifdef DEBUG
	printf("[file_win32] closed the file: status %d\n", ret);
#endif

	// The pipe will be closed, by master.
	node->status = ST_DISCONNECT;
	node->type = LION_TYPE_PIPE_FILE; // Is there a danger master see's this, before we get to exit?

	//printf("[lion_filewriter] signalling %p to be released\n", node);

	//io_force_loop = 1;
	// Alas, we can not to force loop here, since that is OUR force loop, not parents.
	// So, we close the connection instead, this should wake the parent up.
	temp = node->socket;
	node->socket = -1; // Avoid the race condition.
	sockets_close(temp);

	return ret;
}





connection_t *lion_open(char *file, int flags, mode_t modes,
						int lion_flags, void *user_data)
{
	connection_t *newd, *parent = NULL;

	// This breaks FULLFILL, but then, passing NULL breaks lion API.
	if (!file)
		return NULL;
	
	newd = connections_new();
	
	newd->type = LION_TYPE_FILE;
	newd->user_data = user_data;

	if (lion_flags & LION_FLAG_TRACE)
		lion_enable_trace(newd);

	if (newd->trace)
		fprintf(trace_file, "%p: lion_open(%s)\n", newd, file);

	// Open the file, if issues, send event.

	if (lion_flags & LION_FLAG_EXCLUSIVE) {

		newd->socket = sopen(file, 
							 flags | O_BINARY,
							 _SH_DENYWR,
							 modes);

	} else {
		
		newd->socket = open( file, flags | O_BINARY, modes );

	}

#ifdef DEBUG
	if (newd->socket < 0)
		perror("open");
	printf("net_open(%s): %d\n", file, newd->socket);
#endif

	if (newd->socket < 0) {

		if (lion_flags & LION_FLAG_FULFILL) {

			// Signal above layer of failure
			newd->status = ST_PENDING;

			io_force_loop = 1;
			return newd;
		}
	
		newd->status = ST_DISCONNECT;

		return NULL;
	}

	

	// So we now know the file was opened successfully. We now start
	// of the thread(s) based on file open modes.
	// But we have TWO nodes now, One pure FILE that has the fd.
	// and parent gets one end of the pipe.
	if (flags & O_RDWR) {  // Worst case


	} else if (flags & O_WRONLY) {

		parent = lion_filethread(newd->socket, lion_filewriter, NULL, user_data);

	} else {

		//printf("Threading off reader process\n");
		parent = lion_filethread(newd->socket, lion_filereader, NULL, user_data);


	}


	if (!parent) {
		newd->status = ST_PENDING;
		close(newd->socket);
		newd->socket = -1;
		return NULL;
	}

	// Release file node
	parent->file_socket = newd->socket;
	parent->binary = newd->binary;
	parent->user_data = user_data;
	//printf("Node %p has user_data %p\n", parent, user_data);
	newd->socket = -1;
	newd->status = ST_DISCONNECT;  // will be released eventually.




	return parent;

}


#endif
